/* (C) Copyright 1989 Artisoft Inc. */
/*	This program may be distributed free without modification */

/*

   SCOPY - Sharable COPY DOS file program

      Copies files in shared read mode to allow simultaneous copying
      of a file by multiple users using SCOPY. 


      V1.0 - Written - June 8, 1989  LNS

      Modification History
		--------------------

	   V1.04  11/20/89 - Update error messages and comments
      V1.03   8/25/89 - Ignore leading and trailing white space in indirect
                        file names.
      V1.02   8/25/89 - Added comments to indirect file & require only source
      V1.01   8/24/89 - Added @file and check_space().

*/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <types.h>
#include <stat.h>
#include <share.h>
#include <io.h>
#include <dos.h>
#include <ctype.h>

#define buffer_size 30000

char buffer [buffer_size];

char arg1[81], arg2[81], iline[81], batch[81];

char pdrive[_MAX_DRIVE], idrive[_MAX_DRIVE], odrive[_MAX_DRIVE];
char pdir[_MAX_DIR],     idir[_MAX_DIR],     odir[_MAX_DIR];
char pfname[_MAX_FNAME], ifname[_MAX_FNAME], ofname[_MAX_FNAME];
char pext[_MAX_EXT],     iext[_MAX_EXT],     oext[_MAX_EXT];

char rdrive[_MAX_DRIVE], wdrive[_MAX_DRIVE];
char rdir[_MAX_DIR],     wdir[_MAX_DIR];
char rfname[_MAX_FNAME], wfname[_MAX_FNAME];
char rext[_MAX_EXT],     wext[_MAX_EXT];

char ipath[_MAX_PATH], opath[_MAX_PATH];
char rpath[_MAX_PATH], wpath[_MAX_PATH];

char tmpdir[_MAX_DIR + 9];

int rfh, wfh;

FILE *bfile;
int batch_flag;

unsigned bytes_read, bytes_written, file_attrib, w_dev_num;

long total_written, files_copied, fblocks, w_dev_blks_free, old_size;

struct find_t i_file, w_file;

char copyright[50] = "SCOPY V1.04 - (C) Copyright 1989 ARTISOFT Inc.";


/* Main program, gets command line via aguments */

main (int argc, char **argv)
{

/* Get the source file_spec and output directory_spec from command line */

   fprintf (stderr, "%s\n", copyright);

   if (argc == 1)
      {
      fprintf (stderr, "\n");
      fprintf (stderr,
         "Usage is: SCOPY source_file(s)  destination_directory_spec\n");
      fprintf (stderr,
         "          SCOPY @indirect_file  destination_directory_spec\n");
      fprintf (stderr,
         "          Comment characters in indirect file can be !#;'\n");
      exit (1);
      }

   if (argc < 2 || argc > 3)
      {
      fprintf (stderr, "SCOPY-Error: Incorrect number of arguments\n");
      exit (1);
      }

/* Find out where we were run from */

   _splitpath (argv[0], pdrive, pdir, pfname, pext);

/* Copy command line argument(s) to local strings */

   strcpy (arg1, argv[1]);

   if (argc == 3)
      strcpy (arg2, argv[2]);
   else
      arg2[0] = 0;

/* Check first argument for indirect batch file designation */

   strcpy (batch, argv[1]);
   check_batch();

/* Ok, lets get started */

   printf ("\nSCOPYing...\n");
   files_copied = 0;

/* loop here until all source file specifications have been read */

   while (get_file() != 0)

   {
   strupr (arg1);
   strupr (arg2);

   _splitpath (arg1, idrive, idir, ifname, iext);
   _makepath (ipath, idrive, idir, ifname, iext);

   _splitpath (arg2, odrive, odir, ofname, oext);
   _makepath (opath, odrive, odir, ofname, oext);

/* Assume last file name is really directory spec on output side */

   if (odir[0] != NULL && ofname[0] != NULL && oext[0] == NULL &&
       strlen(odir) + strlen(ofname) + 1 <= _MAX_DIR)
      {
      strncat (odir, ofname, 8);
      strncat (odir, "\\", 1);
      ofname[0] = NULL;
      }

   if (oext[0] != NULL || ofname[0] != NULL)
      {
      fprintf (stderr, "SCOPY-Error: Destination file name not allowed\n");
      exit (1);
      }

   if (_dos_findfirst (ipath, _A_NORMAL, &i_file) != 0)
      {
      fprintf (stderr, "SCOPY-Source file '%s'\n", ipath);
      perror ("SCOPY-Error: Invalid source file specification");
      exit (1);
      }

/* Start the file lookups and creates */

   do
      {
      total_written = 0;
      fblocks = 0;

      _splitpath (i_file.name, rdrive, rdir, rfname, rext);
      _makepath (rpath, idrive, idir, rfname, rext);
      _makepath (wpath, odrive, odir, rfname, rext);

      if (strcmp (rpath, wpath) == 0)
         {
         fprintf (stderr,"    %s to %s\n", rpath, wpath);  
         fprintf (stderr,"SCOPY-Error: Cannot copy a file to itself\n");
         exit (1);
         }

/* Does there appear to be enough space to copy the file */

      check_space(odrive[0]);

      printf ("   %s", rpath);

/* Open the input file */

      if (_dos_open (rpath, SH_DENYWR, &rfh) != 0)
         {
         perror ("SCOPY-Error: Open failure on source file");
         exit (1);
         }

      printf ("  to  %s\n", wpath);

/* Create output file; save file attributes across copy */

      file_attrib = i_file.attrib;

      if (_dos_creat (wpath, file_attrib, &wfh) != 0)
         {
         perror ("SCOPY-Error: Open failure on destination file");
         exit (1);
         }

/* Perform copy of indiviual file; buffer by buffer */

      do
         {
          
         if (_dos_read (rfh, buffer, buffer_size, &bytes_read) != 0)
            {
            perror ("SCOPY-Error: Read error");
            exit (1);
            }

         if (_dos_write (wfh, buffer, bytes_read, &bytes_written) != 0)
            {
            perror ("SCOPY-Error: Write error");
            exit (1);
            }

         if (bytes_read != bytes_written)
            {
            fprintf(stderr,"SCOPY-Error: READ/WRITE mismatch (Out of Space)\n");
            exit (1);
            }

         total_written += bytes_written;

/* If a "big" file, then tell user how we are doing */
         
         fblocks++;

         if (fblocks > 2 )
            {
            printf ("   %ld%%\r", 100 * total_written / i_file.size);
            }
         }
      while (bytes_read != 0);

/* Save original file date and time */

      if (_dos_setftime (wfh, i_file.wr_date, i_file.wr_time) != 0)
         {
         perror ("SCOPY-Error: Cannot update date/time on output file");
         exit (1);
         }

/* Close input and output file */

      if (_dos_close (wfh) != 0)
         {
         perror ("SCOPY-Error: Close error on destination file");
         exit (1);
         }

      if (_dos_close (rfh) != 0)
         {
         perror ("SCOPY-Error: Close error on source file");
         exit (1);
         }

      files_copied++;

      }
   while (_dos_findnext (&i_file) == 0);

   }

   printf ("SCOPY copied %ld file(s)\n", files_copied);
   exit (0);

}


/* Check for sufficient free space on disk to copy file */

check_space(char drive_letter)
{
   int flag, fflag;
   long old_space;

   do
      {
      flag = 0;

/* get free disk space on output device */

      toupper (drive_letter);

      if (drive_letter < 'A' || drive_letter > 'Z')
         w_dev_num = 0;
      else
         w_dev_num = drive_letter - 'A' + 1;

      get_dev_free (w_dev_num, &w_dev_blks_free);

/* Check if file exists on output device and if so get its size */

      fflag = _dos_findfirst (wpath, _A_NORMAL, &w_file);

      if (fflag == 0)
         old_space = w_file.size;
      else
         old_space = 0;

/* If new file will not fit into free space combined with its old size */

      if (i_file.size > w_dev_blks_free + old_space)
         {
         flag = -1;

/* Tell user about old file on output device */

         if (fflag == 0)
            {
            fprintf (stderr,
               "\nSCOPY-Warning: Output file '%s' not overwritten\n", wpath);
            }

/* See what user wants to do about the problem */

         do
            {
            fprintf (stderr, "\n");
            fprintf (stderr,
               "File '%s' will not fit on output device.\n", rpath);

            fprintf (stderr,
               "Blocks Available: %ld    ", w_dev_blks_free+old_space);

            fprintf (stderr, "Blocks Needed: %ld\n", i_file.size);
            fprintf (stderr, "A-bort SCOPY or R-etry with new media [A,R]: ");
            strupr (gets (iline));
            }
         while (iline[0] != 'A' & iline[0] != 'R');

         }

      if (iline[0] == 'A')
         exit (1);

      }
   while (flag != 0);

}


/* Calculate number of blocks free on a device */

get_dev_free (unsigned dev_num, long *bytes_free)
   {
   struct diskfree_t drive;

   if (_dos_getdiskfree (dev_num, &drive) == 0)
      {
      *bytes_free = (long) drive.avail_clusters *
                    (long) drive.sectors_per_cluster *
                    (long) drive.bytes_per_sector;
      }
   else
      {
      perror ("SCOPY-Error: Can not access destination device");
      exit (1);
      }

   return;

   }


/* check for presence of batch indirect file */

check_batch ()
{
   strupr (batch);

/* if indirect file specified, then set batch flag to -1, else batch flag = 1 */

   if (batch[0] == '@')
      batch_flag = -1;
   else
      {
      batch_flag = 1;
      return;
      }

/* open the indirect batch file if need be */

   if ((bfile = fopen(&batch[1],"r")) == NULL)
      {
      fprintf (stderr, "SCOPY-Error: Indirect file '%s'\n", &batch[1]);
      perror ("Could not open indirect file");
      exit (1);              
      }

}


/* get the next input file; either from the command line or indirect file */

int get_file()
{
   int i;

/* are we in batch mode with indirect file or single file from command line */

   if (batch_flag > 0)
      {
      batch_flag = 0;
      return (1);
      }
   else if (batch_flag == 0)
      {
      return (0);
      }


   while (fgets(iline, 80, bfile) != NULL)
      {

/* trim leading white space */

      for (i=0; i < strlen(iline) && iline[i] <= ' '; i++);

      strcpy (arg1, &iline[i]);

/* trim trailing white space */

      for (i=strlen(arg1); i >= 0 && arg1[i] <= ' '; i--)
         arg1[i] = 0;

/* Omit comment and blank lines */

      if (arg1[0] == 0 || strchr ("!#;'", arg1[0]) != NULL)
         continue;

      return (1);
      }

   return (0);
}

/* (C) Copyright 1989 Artisoft Inc. */

